//
//  NEScreenShareHost.m
//  NEScreenShareKit
//
//  Created by Wenchao Ding on 2021/5/13.
//  Copyright © 2021 NetEase. All rights reserved.
//

#import "NEScreenShareHost.h"
#import <ReplayKit/ReplayKit.h>
#import "NEScreenShareHostPrivateDefine.h"
#import "NEScreenShareHostProtocol.h"
#import "GCDAsyncSocket.h"

static void *KVOContext = &KVOContext;

static void onBroadcastStarted(CFNotificationCenterRef center,
                               void *observer,
                               CFStringRef name,
                               const void *object,
                               CFDictionaryRef
                               userInfo) {
    NEScreenShareHost *host = (__bridge NEScreenShareHost *)(observer);
    id<NEScreenShareHostDelegate> delegate = [[host valueForKey:@"options"] delegate];
    if (delegate && [delegate respondsToSelector:@selector(onBroadcastStarted)]) {
        [delegate onBroadcastStarted];
    }
}

static void onBroadcastFinished(CFNotificationCenterRef center,
                                void *observer,
                                CFStringRef name,
                                const void *object,
                                CFDictionaryRef
                                userInfo) {
    NEScreenShareHost *host = (__bridge NEScreenShareHost *)(observer);
    id<NEScreenShareHostDelegate> delegate = [[host valueForKey:@"options"] delegate];
    if (delegate && [delegate respondsToSelector:@selector(onBroadcastFinished)]) {
        [delegate onBroadcastFinished];
    }
}

static void onBroadcastPaused(CFNotificationCenterRef center,
                              void *observer,
                              CFStringRef name,
                              const void *object,
                              CFDictionaryRef
                              userInfo) {
    NEScreenShareHost *host = (__bridge NEScreenShareHost *)(observer);
    id<NEScreenShareHostDelegate> delegate = [[host valueForKey:@"options"] delegate];
    if (delegate && [delegate respondsToSelector:@selector(onBroadcastPaused)]) {
        [delegate onBroadcastPaused];
    }
}

static void onBroadcastResumed(CFNotificationCenterRef center,
                               void *observer,
                               CFStringRef name,
                               const void *object,
                               CFDictionaryRef
                               userInfo) {
    NEScreenShareHost *host = (__bridge NEScreenShareHost *)(observer);
    id<NEScreenShareHostDelegate> delegate = [[host valueForKey:@"options"] delegate];
    if (delegate && [delegate respondsToSelector:@selector(onBroadcastResumed)]) {
        [delegate onBroadcastResumed];
    }
}


@interface NEScreenShareHost () <NEScreenShareHostProtocol, GCDAsyncSocketDelegate>

@property (nonatomic, strong) NEScreenShareHostOptions *options;
@property (nonatomic, strong) NSUserDefaults *userDefaults;
@property (nonatomic, strong) RPSystemBroadcastPickerView *systemPicker API_AVAILABLE(ios(12.0));

@property (nonatomic, strong) GCDAsyncSocket *serverSocket;
@property (nonatomic, assign) int listenPort;
@property (nonatomic, strong) NSMutableArray *clientSocket;


@property (nonatomic, assign) NSInteger timeout;
@property (nonatomic, strong) NSNumber *sizeNumber;
@property (nonatomic, strong) NSMutableDictionary *parameter;

@end

@implementation NEScreenShareHost

- (instancetype)initWithOptions:(nonnull NEScreenShareHostOptions *)options {
    self = [super init];
    if (self) {
        NSAssert(options.appGroup.length, @"请输入合理的appGroup");
        NSAssert(options.delegate, @"请添加delegate");
        self.options = options;
        
        [self startListenServer];
        
        self.userDefaults = [[NSUserDefaults alloc] initWithSuiteName:options.appGroup];
        
        [self.userDefaults addObserver:self forKeyPath:NEScreenShareVideoFrameKey options:NSKeyValueObservingOptionNew context:KVOContext];
        CFNotificationCenterAddObserver(CFNotificationCenterGetDarwinNotifyCenter(),
                                        (__bridge const void *)(self),
                                        onBroadcastStarted,
                                        (__bridge CFStringRef)NEScreenShareBroadcastStartedNotification,
                                        NULL,
                                        CFNotificationSuspensionBehaviorDeliverImmediately);
        CFNotificationCenterAddObserver(CFNotificationCenterGetDarwinNotifyCenter(),
                                        (__bridge const void *)(self),
                                        onBroadcastFinished,
                                        (__bridge CFStringRef)NEScreenShareBroadcastFinishedNotification,
                                        NULL,
                                        CFNotificationSuspensionBehaviorDeliverImmediately);
        CFNotificationCenterAddObserver(CFNotificationCenterGetDarwinNotifyCenter(),
                                        (__bridge const void *)(self),
                                        onBroadcastPaused,
                                        (__bridge CFStringRef)NEScreenShareBroadcastPausedNotification,
                                        NULL,
                                        CFNotificationSuspensionBehaviorDeliverImmediately);
        CFNotificationCenterAddObserver(CFNotificationCenterGetDarwinNotifyCenter(),
                                        (__bridge const void *)(self),
                                        onBroadcastResumed,
                                        (__bridge CFStringRef)NEScreenShareBroadcastResumedNotification,
                                        NULL,
                                        CFNotificationSuspensionBehaviorDeliverImmediately);
    }
    return self;
}

- (void)dealloc {
    [self.userDefaults removeObserver:self forKeyPath:NEScreenShareVideoFrameKey];
    CFNotificationCenterRemoveObserver(CFNotificationCenterGetDarwinNotifyCenter(), (__bridge const void *)(self), (__bridge CFStringRef)NEScreenShareBroadcastStartedNotification, NULL);
    CFNotificationCenterRemoveObserver(CFNotificationCenterGetDarwinNotifyCenter(), (__bridge const void *)(self), (__bridge CFStringRef)NEScreenShareBroadcastFinishedNotification, NULL);
    CFNotificationCenterRemoveObserver(CFNotificationCenterGetDarwinNotifyCenter(), (__bridge const void *)(self), (__bridge CFStringRef)NEScreenShareBroadcastPausedNotification, NULL);
    CFNotificationCenterRemoveObserver(CFNotificationCenterGetDarwinNotifyCenter(), (__bridge const void *)(self), (__bridge CFStringRef)NEScreenShareBroadcastResumedNotification, NULL);
    
}

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary<NSKeyValueChangeKey,id> *)change context:(void *)context {
    if (context != KVOContext) {
        return [super observeValueForKeyPath:keyPath ofObject:object change:change context:context];
    }
    if ([keyPath isEqualToString:NEScreenShareVideoFrameKey] && self.options.delegate) {
        NSDictionary *i420Frame = change[NSKeyValueChangeNewKey];
        if ([self.options.delegate respondsToSelector:@selector(onReceiveVideoFrameInfo:)]) {
            [self.options.delegate onReceiveVideoFrameInfo:i420Frame];
        }
    }
}

- (void)launchBroadcasterWithPort:(int)port API_AVAILABLE(ios(12.0)) {
    
    NSString *plugInPath = NSBundle.mainBundle.builtInPlugInsPath;
    if (!plugInPath) {
        return;
    }
    NSArray *contents = [NSFileManager.defaultManager contentsOfDirectoryAtPath:plugInPath error:nil];
    for (NSString *content in contents) {
        NSURL *url = [NSURL fileURLWithPath:plugInPath];
        NSBundle *bundle = [NSBundle bundleWithPath:[url URLByAppendingPathComponent:content].path];
        
        NSDictionary *extension = [bundle.infoDictionary objectForKey:@"NSExtension"];
        if (extension == nil) { continue; }
        NSString *identifier = [extension objectForKey:@"NSExtensionPointIdentifier"];
        if ([identifier isEqualToString:@"com.apple.broadcast-services-upload"]) {
            self.systemPicker.preferredExtension = bundle.bundleIdentifier;
            break;
        }
    }
    for (UIView *view in self.systemPicker.subviews) {
        UIButton *button = (UIButton *)view;
        [button sendActionsForControlEvents:UIControlEventAllTouchEvents];
    }
}

- (void)launchBroadcaster API_AVAILABLE(ios(12.0)) {
    [self launchBroadcasterWithPort:self.options.listenPort];
}

- (void)stopBroadcaster {
    CFNotificationName notificationName = CFNotificationName(NEScreenShareHostRequestStopNotification);
    CFNotificationCenterPostNotification(CFNotificationCenterGetDarwinNotifyCenter(), notificationName, nil, nil, true);
}


- (RPSystemBroadcastPickerView *)systemPicker {
    if (!_systemPicker) {
        RPSystemBroadcastPickerView* picker =
        [[RPSystemBroadcastPickerView alloc] initWithFrame:CGRectMake(0, 0, 44, 44)];
        picker.showsMicrophoneButton = NO;
        picker.autoresizingMask = UIViewAutoresizingFlexibleTopMargin | UIViewAutoresizingFlexibleRightMargin;
        _systemPicker = picker;
    }
    return _systemPicker;
}

-(void)startListenServer{
    self.timeout = -1;
    self.clientSocket = [[NSMutableArray alloc] init];
    self.serverSocket = [[GCDAsyncSocket alloc] initWithDelegate:self delegateQueue:dispatch_queue_create("data_queue", NULL)];
}

- (NSError *)prepare {
    NSError *err;
    if (self.listenPort == 0) {
        [self.serverSocket disconnect];
        //打开监听端口
        [self.serverSocket acceptOnInterface:@"localhost" port:self.options.listenPort error:&err];
        if (!err) {
            self.listenPort = self.serverSocket.localPort;
            NSLog(@"NEScreenShareHost server start listening success. port: %d %d", self.options.listenPort, self.listenPort);
        } else{
            NSLog(@"NEScreenShareHost server start listening error. code: %@, desc: %@, reason: %@", @(err.code), err.localizedDescription, err.localizedFailureReason);
        }
    }
    
    [self.userDefaults setObject:[NSNumber numberWithInt:self.listenPort] forKey:@"port"];
    [self.userDefaults synchronize];
    return err;
}

#pragma mark 有客户端建立连接的时候调用
-(void)socket:(GCDAsyncSocket *)sock didAcceptNewSocket:(GCDAsyncSocket *)newSocket{
    NSLog(@"NEScreenShareHost didAcceptNewSocket");
    [self.clientSocket addObject:newSocket];
    [newSocket readDataToData:GCDAsyncSocket.CRLFData withTimeout:self.timeout tag:0];
}

#pragma mark 服务器写数据给客户端
-(void)socket:(GCDAsyncSocket *)sock didWriteDataWithTag:(long)tag{
    NSLog(@"NEScreenShareHost Server didWriteDataWithTag %s",__func__);
    [sock readDataWithTimeout:self.timeout tag:tag];
}

#pragma mark 接收客户端传递过来的数据
-(void)socket:(GCDAsyncSocket *)sock didReadData:(NSData *)data withTag:(long)tag{
    //sock为客户端的socket
    NSLog(@"NEScreenShareHost server socket did receive data :  %ld", data.length);
    NSLog(@"NEScreenShareHost data tag : %ld",tag);

    // only local socket
    if (self.sizeNumber == nil) {
        NSError *error = nil;
        NSData *numberData = [data subdataWithRange:NSMakeRange(0, data.length - GCDAsyncSocket.CRLFData.length)];
        NSLog(@"NEScreenShareHost number data size : %lu",numberData.length);
        if (@available(iOS 11.0, *)) {
            self.sizeNumber = [NSKeyedUnarchiver unarchivedObjectOfClass:[NSNumber class] fromData:numberData error:&error];
        }
        if (error) {
            NSLog(@"NEScreenShareHost parse number error : %@",error);
            return;
        }
        NSLog(@"NEScreenShareHost data length : %d",self.sizeNumber.intValue);
        [sock readDataToLength:self.sizeNumber.integerValue withTimeout:self.timeout tag:tag];
        return;
    }
    
    if (self.sizeNumber == nil ||  self.sizeNumber.integerValue != data.length) {
        NSLog(@"NEScreenShareHost error: packet size is wrong");
        return;
    }
    
    NSError *error;
    if (@available(iOS 11.0, *)) {
        NSDictionary *i420Frame = [NSKeyedUnarchiver unarchivedObjectOfClass:[NSMutableDictionary class] fromData:data error:&error];
        if (error == nil) {
            [self.options.delegate onReceiveVideoFrameInfo:i420Frame];
            NSLog(@"NEScreenShareHost send yuv data ");
        }else {
            NSLog(@"NEScreenShareHost convert yuv error : %@",error);
        }
    }
    self.sizeNumber = nil;
    [sock readDataToData:[GCDAsyncSocket CRLFData] withTimeout:self.timeout tag:tag];
    
}

@end
